# This file is part of the Falcon repository creator
# Copyright (C) 2005-2006 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import falcon.conf
import falcon.util
import falcon.pkglist
import falcon.html
import shutil
import os

def update(mirror, release):
    if not falcon.conf.pretend: 
        # Remove components that no longer exist
        if os.path.exists(os.path.join(mirror['basedir'],'dists',release)):
            for oldsect in os.listdir(os.path.join(mirror['basedir'],'dists',release)):
                if os.path.isdir(os.path.join(mirror['basedir'],'dists',release,oldsect)):
                    if oldsect not in falcon.conf.components[release] or \
                        (oldsect not in mirror['components'] and mirror['components'] != ['all']):
                        shutil.rmtree(os.path.join(mirror['basedir'],'dists',release,oldsect))
        #if not os.path.exists(os.path.join(mirror['basedir'],'dists',release)):
        else:
            os.makedirs(os.path.join(mirror['basedir'],'dists',release))
        if mirror['components'] == ['all']:
            copydir(os.path.join(falcon.conf.basedir,'dists',release),
                    os.path.join(mirror['basedir'],'dists',release))
        else:
            for sect in falcon.util.intersect(mirror['components'], falcon.conf.components[release]):
                if not os.path.exists(os.path.join(mirror['basedir'],'dists',release,sect)):
                    os.mkdir(os.path.join(mirror['basedir'],'dists',release,sect))
                copydir(os.path.join(falcon.conf.basedir,'dists',release, sect),
                        os.path.join(mirror['basedir'],'dists',release, sect))
            falcon.pkglist.release(os.path.join(mirror['basedir'],'dists',release))

def copydir(d1, d2):
    falcon.util.debug("Recursively copying %s to %s" % (d1, d2))
    for f in os.listdir(d1):
        if os.path.isdir(os.path.join(d1,f)):
            if not os.path.exists(os.path.join(d2,f)): 
                os.mkdir(os.path.join(d2, f))
            copydir(os.path.join(d1,f), os.path.join(d2,f))
        else:
            if falcon.util.newer(os.path.join(d1,f), os.path.join(d2,f)):
                shutil.copy(os.path.join(d1,f), os.path.join(d2,f))

def html(mirror, release):
    if not os.path.exists(mirror['basedir']):
        falcon.util.warning("Not updating html for %s, basedir %s does not exist." % (mirror['alias'], mirror['basedir']))
        falcon.util.warning("Forgot to run falcon update?")
        return
    if not falcon.conf.pretend:
        for f in falcon.conf.webfiles:
            if os.path.exists(os.path.join(falcon.conf.basedir, f)):
                if falcon.util.newer(os.path.join(falcon.conf.basedir, f), os.path.join(mirror['basedir'],f)):
                    shutil.copy(os.path.join(falcon.conf.basedir, f), mirror['basedir'])
    falcon.html.generate(mirror['basedir'], mirror) # Preferably this is moved outside the loop
    falcon.html.generate(os.path.join(mirror['basedir'],'dists',release), mirror)

    for component in falcon.util.intersect(falcon.conf.components[release], mirror['components']):
        falcon.html.generate(os.path.join(mirror['basedir'], 'dists', release, component), mirror)

@falcon.util.timer
def sync(mirror):
    # Check for the existence of the basedir
    if not os.path.exists(mirror['basedir']):
        falcon.util.warning("Not syncing with %s, basedir %s does not exist." % (mirror['alias'], mirror['basedir']))
        falcon.util.warning("Forgot to run falcon update?")
        return
    args = '-ruRlt'
    if falcon.conf.pretend: args += 'n'
    if falcon.conf.verbose: args += ' --progress'
    if mirror['rsync_opts']:
        args += ' %s' % mirror['rsync_opts']
    basedir = os.path.join(falcon.conf.basedir, 'base-%s' % mirror['alias'])
    rfilter =  '''--filter '+ **deb' --filter '+ **dsc' --filter '+ **tar.gz' '''
    rfilter += '''--filter '+ **diff.gz' --filter '+ **changes' --filter '- **' '''

    # Which folders to upload?
    # - pool/$rel/$sec if $sec in m[components]
    # - pool/$sec/ if symlink to pool/$sec in previous (to support that hack)
    dirs = []
    for r in falcon.util.intersect(falcon.conf.releases, mirror['releases']):
        dirs.append(os.path.join('pool',r))
        for s in falcon.util.intersect(falcon.conf.components[r], mirror['components']):
            if s == 'all':
                continue
            dirs.append(os.path.join('pool',r,s))
            # This is the hack
            if os.path.islink(os.path.join(falcon.conf.basedir, 'pool',r,s)):
                dirs.append(os.path.normpath(os.path.join('pool',r,os.readlink(os.path.join(falcon.conf.basedir, 'pool',r,s)))))
    fdirs = " ".join(["--filter '+ %s'" % x for x in dirs])
    dirs = " ".join(dirs)

    rfilter = "--filter '+ pool/' " + fdirs + ' ' + rfilter
    # Upload data
    try:
        falcon.util.run("rsync %s %s %s '%s'" %  ( args, rfilter, dirs, mirror['rsync']), wd = falcon.conf.basedir)
    except:
        falcon.util.warning("Synchronizing with %s failed" % mirror['alias'])
        return
    # Upload meta
    try:
        falcon.util.run("rsync %s --delete --force * '%s'" % (args, mirror['rsync']), wd=mirror['basedir'])
    except:
        falcon.util.warning("Synchronizing with %s failed" % mirror['alias'])
        return
    # Sync data, now with --delete :)
    try:
        falcon.util.run("rsync %s --existing --delete --force %s '%s'" % (args, 'pool', mirror['rsync']), wd=falcon.conf.basedir)
    except:
        falcon.util.warning("Synchronizing with %s failed" % mirror['alias'])
        return
