# This file is part of the Falcon repository creator
# Copyright (C) 2005-2006 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import falcon.conf
import falcon.util
import os, time
import re

_sect_re = re.compile(r'(\nSection: )(?:.*?/)?(.*)')

_release_template = """\
Origin: %s
Label: %s
Suite: %s
Version: %s
Codename: %s
Architectures: %s
Components: %s
Description: %s
%s
"""

_sect_release_template = """\
Archive: %s
Version: %s
Component: %s
Origin: %s
Label: %s
Architecture: %%s
"""

@falcon.util.timer
def update(release, component):
    sect_dir     = os.path.join(falcon.conf.basedir, 'dists', release, component)
    
    # Get apt-ftparchive output
    if falcon.conf.do_all and component == 'all':
        if len(falcon.conf.components[release]) == 1:
            falcon.util.warning('Not creating all component, no components exist')
            return
        elif len(falcon.conf.components[release]) == 2:
            sects = falcon.conf.components[release][0]
        else:
            sects = '{%s}' % ",".join([x for x in falcon.conf.components[release] if x != 'all'])
        sources = falcon.util.run("cat %s" % os.path.join(falcon.conf.basedir, 'dists', release, '%s' % sects,
                                                                'source','Sources'))
        packages = falcon.util.run("cat %s" % os.path.join(falcon.conf.basedir, 'dists', release, '%s' % sects,
                                                                 'binary-*','Packages'))
    else:
        sources = falcon.util.run("apt-ftparchive sources %s" % os.path.join('pool',release,component),
                                        wd = falcon.conf.basedir).strip() + "\n\n"
        packages = falcon.util.run("apt-ftparchive --db falcon.db packages %s" % os.path.join('pool',release,component),
                                         wd = falcon.conf.basedir)
        # Add the component
        #packages = packages.replace('\nSection: ', '\nSection: %s/' % component)
        packages = _sect_re.sub(r"\g<1>%s/\g<2>" % re.escape(component), packages)
    # Split per architecture
    allpackages = [p.strip() for p in packages.split('\n\n') if p.strip()]
    packages = dict(zip(falcon.conf.architectures, ['']*len(falcon.conf.architectures)))
    for p in allpackages:
        pos = p.find('\nArchitecture: ') + 15
        if p[pos:pos+3] == 'all':
            for a in packages:
                packages[a] += p + '\n\n'
        else:
            for a in packages:
                if p[pos:pos+len(a)+1] == a + '\n':
                    packages[a] += p + '\n\n'
                    break
            else:
                arch = p[pos:]
                arch = arch[:arch.find("\n")]
                p = p[p.find('\nPackage: ')+10:]
                p = p[:p.find("\n")]
                falcon.util.error("Package %s in component %s has unknown architecture %s" % (p, component, arch ))

    # Now write it all
    if not falcon.conf.pretend:
        releasedata = _sect_release_template % (release, falcon.conf.version, component, 
                                                falcon.conf.origin, falcon.conf.label)
        if sources:
            sourcefile   = os.path.join(sect_dir, 'source', 'Sources')
            if not os.path.exists(os.path.join(sect_dir, 'source')):
                os.makedirs(os.path.join(sect_dir, 'source'))
            falcon.util.writefile(sourcefile, sources)
            if falcon.util.newer(sourcefile, sourcefile + '.gz'):
                falcon.util.run("gzip < '%s' > '%s'.gz"   % (sourcefile,  sourcefile))
                falcon.util.run("bzip2 < '%s' > '%s'.bz2" % (sourcefile,  sourcefile))
            falcon.util.writefile(os.path.join(sect_dir, 'source', 'Release'), releasedata % 'source')
        else:
            falcon.util.warning("Component %s contains no sources, this may confuse package managers" % component)
        for arch in packages:
            if packages[arch]:
                packagefile  = os.path.join(sect_dir, 'binary-%s' % arch, 'Packages')
                if not os.path.exists(os.path.join(sect_dir, 'binary-%s' % arch)):
                    os.makedirs(os.path.join(sect_dir, 'binary-%s' % arch))
                falcon.util.writefile(packagefile, packages[arch])
                if falcon.util.newer(packagefile, packagefile + '.gz'):
                    falcon.util.run("gzip < '%s' > '%s'.gz"   % (packagefile, packagefile))
                    falcon.util.run("bzip2 < '%s' > '%s'.bz2" % (packagefile, packagefile))
                falcon.util.writefile(os.path.join(sect_dir, 'binary-%s' % arch, 'Release'), releasedata % arch)
            else:
                falcon.util.warning("Component %s contains no packages for %s, this may confuse package managers" % (component, arch))

@falcon.util.timer
def release(dir):
    relfile = os.path.join(dir, 'Release')
    rel = os.path.basename(dir)
    if not os.path.exists(dir):
        os.makedirs(dir)
    reldata = falcon.util.run("apt-ftparchive release .", wd = dir) 
    reldata = _release_template % (falcon.conf.origin,  falcon.conf.label, rel, falcon.conf.version, rel, 
                                   " ".join(falcon.conf.architectures), " ".join(falcon.conf.components[rel]),
                                   falcon.conf.description, reldata)
    falcon.util.writefile(relfile, reldata)

    if not falcon.conf.gpgkey:
        falcon.util.warning("No signing key available. Your repository will be unauthenticated")
    else:
        falcon.util.debug("Signing release file for %s" % dir)
        if not falcon.conf.pretend and falcon.util.newer(relfile, relfile + '.gpg'):
            if os.path.exists(relfile + '.gpg'):
                os.unlink(relfile + '.gpg')
            falcon.util.run("gpg -u %s --detach-sig --armor --output %s.gpg %s" % (falcon.conf.gpgkey, relfile, relfile))
