# This file is part of the Falcon repository creator
# Copyright (C) 2005-2006 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import os
from email import FeedParser
import warnings
warnings.filterwarnings("ignore", "apt API not stable yet", FutureWarning)
import apt
import re
import stat

import falcon.util
import falcon.conf
import falcon.html

_unique = 0
cache = {}
@falcon.util.timer
def scan_sources(sourcedir):
    """ Scan a dists/dist/ subdir """
    global _unique, cache
    if sourcedir in cache:
        return cache[sourcedir]
    pdata = []
    sdata = []
    
    if len(falcon.util.run('ls %s' % os.path.join(sourcedir, 'binary-*', 'Packages'))):
        pdata = falcon.util.run('cat %s' % os.path.join(sourcedir, 'binary-*', 'Packages')).split('\n\n')
    if os.path.exists(os.path.join(sourcedir, 'source', 'Sources')):
        sdata = falcon.util.readfile(os.path.join(sourcedir, 'source', 'Sources'),
                                     ignore_errors = False).split("\n\n")
    packages = {}
    sources = {}
    # Parse all packages - this will add all packages without Source: to the sources too!
    for p in pdata:
        if not p.strip():
            continue
        f = FeedParser.FeedParser()
        f.feed(p.strip())
        p = f.close()
        desc = p['Description'].replace('&','&amp;').replace('<','&lt;').replace('>','&gt;').split("\n",1)
        if len(desc) > 1:
            desc[1] = '<p>' + desc[1].strip().replace(" .\n",'</p><p>').replace('\n','<br />') + '</p>'
        else:
            desc.append('')
        p['Description2'] = desc
        if '%s_%s' % (p['Package'], p['Version']) in packages:
            packages['%s_%s' % (p['Package'], p['Version'])][p['architecture']] = p
        else:
            packages['%s_%s' % (p['Package'], p['Version'])] = {0:p, p['architecture']: p}
    # Parse sources
    for s in sdata:
        if not s.strip():
            continue
        f = FeedParser.FeedParser()
        f.feed(s.strip())
        s = f.close()
        if s['Package'] not in sources or apt.VersionCompare(s['Version'], sources[s['Package']]['Version']) >= 0:
            s['ffiles'] = map(_l2f,s['Files'].strip().split("\n"))
            _packages = []
            for p in s['Binary'].split(','):
                try:
                    _p = packages.pop('%s_%s' % (p.strip(), s['Version']))
                    _packages.append(_p)
                except KeyError: # Package not built yet?
                    _packages.append({0: {'Package': p,
                                              'Description2': ('Not yet built!','This binary package has not yet been built.'),
                                              'Filename': None,
                                              'Version': _unique}})
                    _unique += 1
            _packages.sort(lambda x, y: x[x.keys()[0]]['Package'] < y[y.keys()[0]]['Package'] and -1 or 1)
            s['packages'] = _packages
            sources[s['Package']] = s
            get_changes(s)
    # packages now should be empty, but m-a and others create packages with 
    # nonexisting sources in the Source: header
    # Also, some packages dont have a Source: and don't have a sourcepackage,
    # like the ones make-jpkg generates
    for p in packages.keys():
        p = packages[p]
        p0 = p[p.keys()[0]]
        if not p0['Source']:
            p0['Source'] = p0['Package']
        source = p0['Source']
        version = p0['version']
        # Remove +.* from the version, module assistant appends the kernel
        # version in this way.
        if '+' in version:
            version = version[:version.rfind('+')]
            if not ('-' in version and '.' in version):
                version = p0['version']
        if source not in sources or apt.VersionCompare(version, sources[source]['Version']) > 0:
            sources[source] = {'Package': source, 'Version': version, 
                               'packages': [p], 'ffiles': [], 'changes': ''}
            get_changes(sources[source])
        else:
            if apt.VersionCompare(version, sources[source]['Version']) == 0:
                # Filter out older binary compatible kernel packages
                addme = True
                for p2 in sources[source]['packages']:
                    if p2[p2.keys()[0]]['package'] == p0['package']:
                        if p2[p2.keys()[0]]['version'] > p0['version']:
                            addme = False
                        else:
                            sources[source]['packages'].remove(p2)
                if addme:
                    sources[source]['packages'].append(p)
                sources[source]['packages'].sort(lambda x, y: x[x.keys()[0]]['Package'] < y[y.keys()[0]]['Package'] and -1 or 1)
                get_changes(sources[source])
    sourcekeys = sources.keys()
    sourcekeys.sort()
    cache[sourcedir] = (sources, sourcekeys)
    return (sources, sourcekeys)

_changes_re = re.compile(r'\nChanges:(.*?)\nFiles',re.DOTALL)
def get_changes(source):
    """ Try to find a changelog for a package """
    # Attempt one: _source.changes
    changelog = ''
    try:
        file = [f for f in source['ffiles'] if f[3] == 'dsc'][0][2].replace('.dsc','_source.changes')
    except:
        pass # Source without files
    else:
        if os.path.exists(os.path.join(falcon.conf.basedir, source['directory'], file)):
            changelog = os.path.join(falcon.conf.basedir, source['directory'], file)
    if not changelog:
        # Attempt 2: All binaries
        for q in source['packages']:
            for p in q:
                p = q[p]
                if not p['Filename']:
                    continue
                if os.path.exists(os.path.join(falcon.conf.basedir, p['Filename'].replace('deb','changes'))):
                    changelog = os.path.join(falcon.conf.basedir, p['Filename'].replace('deb','changes'))
                # Falback to _$arch.changes
                elif '_all.deb' in p['Filename']:
                    for arch in falcon.conf.architectures:
                        if os.path.exists(os.path.join(falcon.conf.basedir, p['Filename'].replace('_all.deb', '_%s.changes' % arch))):
                            changelog = os.path.join(falcon.conf.basedir, p['Filename'].replace('_all.deb','_%s.changes' % arch))
                            break
                if changelog:
                    break

    if changelog:
        source['changelog'] = changelog
        data = falcon.util.readfile(changelog, ignore_errors = False)
        reo = _changes_re.search(data)
        try:
            source['changes'] = falcon.html.format_changelog(reo.group(1))
        except:
            raise
            falcon.util.warning("Invalid data in changelog: %s" % source['changelog'])
    else:
        source['changelog'] = ''
        source['changes'] = ''

def _l2f(l):
    """ From a file line to a [shasum, size, filename, extension] list """
    f = l.strip().split(None,2)
    e = f[2].split('.')
    if e[-1] == 'dsc':    f.append(e[-1])
    elif e[-3] == 'orig': f.append('.'.join(e[-3:]))
    else:                 f.append('.'.join(e[-2:]))
    return f

def clean():
    to_keep = []
    for release in falcon.conf.releases:
        to_keep.append(os.path.join(falcon.conf.basedir, 'pool',release))
        to_keep.append(os.path.join(falcon.conf.basedir, 'pool',release,'desc'))
        to_keep.append(os.path.join(falcon.conf.basedir, 'pool',release,'falcon.ini'))
        for component in falcon.conf.components[release]:
            if component == 'all':
                continue
            sources,sourcekeys = scan_sources(os.path.join(falcon.conf.basedir, 'dists', release, component))
            to_keep.append(os.path.join(falcon.conf.basedir,'pool',release,component))
            to_keep.append(os.path.join(falcon.conf.basedir,'pool',release,component,'desc'))
            for s in sourcekeys:
                for f in sources[s]['ffiles']:
                    to_keep.append(os.path.join(falcon.conf.basedir, sources[s]['directory'], f[2]))
                for q in sources[s]['packages']:
                    for p in q:
                        if p == 0: continue
                        p = q[p]
                        if p['Filename']:
                            to_keep.append(os.path.join(falcon.conf.basedir,p['Filename']))
                if sources[s]['changelog']:
                    to_keep.append(sources[s]['changelog'])

    for f in to_keep:
        # FIXME More symlink magic here *should* work, since all folders are in
        # here too. Make sure to protect against symlink madness
        r = os.path.realpath(f)
        if f != r:
            while r.startswith(os.path.join(falcon.conf.basedir,'pool') + os.path.sep):
                if r not in to_keep:
                    to_keep.append(r)
                r = os.path.dirname(r)

    # Support the symlink compatibility hack here too: only descend into subdirs
    # if falcon.ini exists in the current dir
    existing_files = []
    existing_dirs  = []
    
    dir = os.path.join(falcon.conf.basedir,'pool')
    contents = [os.path.join(dir, x) for x in os.listdir(dir)]
    for name in contents:
        if os.path.isdir(name):
            existing_dirs.append(name)
        else:
            existing_files.append(name)

    # One level deep: releases
    for dir in list(existing_dirs):
        if os.path.exists(os.path.join(dir,'falcon.ini')):
            contents = [os.path.join(dir, x) for x in os.listdir(dir)]
            for name in contents:
                if os.path.isdir(name):
                    # Second level: components
                    contents2 = [os.path.join(dir, name, x) for x in os.listdir(os.path.join(dir,name))]
                    for name2 in contents2:
                        if os.path.isdir(name2):
                            existing_dirs.append(name2)
                        else:
                            existing_files.append(name2)
                else:
                    existing_files.append(name)

    also_keep = []
    if os.path.exists(os.path.join(falcon.conf.basedir,'falcon.keep')):
        also_keep = [x.strip() for x in open(os.path.join(falcon.conf.basedir,'falcon.keep')).readlines()]
        to_keep += also_keep

    to_remove = [x for x in existing_dirs + existing_files if x not in to_keep and os.path.realpath(x) not in to_keep]
    
    if len(to_remove):
        for f in sorted(to_remove):
            # Symlinks can simply be removed
            fname = os.path.basename(f)
            s = os.lstat(f)
            if stat.S_ISLNK(s[stat.ST_MODE]):
                if falcon.util.yesno("Remove obsolete symlink %s?" % f, True):
                    if not falcon.conf.pretend:
                        os.unlink(f)
                else:
                    also_keep.append(f)
            else:
                if not os.path.exists(os.path.join(falcon.conf.basedir, 'morgue', fname)):
                    if falcon.util.yesno("Move obsolete file %s to the morgue?" % f, True):
                        if not falcon.conf.pretend:
                            os.rename(f, os.path.join(falcon.conf.basedir, 'morgue', fname))
                    else:
                        also_keep.append(f)
                else:
                    # Here's where it gets tricky
                    if falcon.util.yesno("morgue/%s already exists. Do you want to delete %s?" % (fname, f), False):
                        if not falcon.conf.pretend:
                            os.unlink(f)
                    else:
                        also_keep.append(f)
        if not falcon.conf.pretend:
            falcon.util.writefile(os.path.join(falcon.conf.basedir, 'falcon.keep'), "\n".join(also_keep))

